
/**************************************************************************************
**
**  The Osiris application compares one catalog of executable files
**  with another catalog of executable files. Osiris catalogs specified
**  directories of files (including MD5 hashes, modification dates,
**  and file attributes into a specified database.
**
**  Copyright (C) 2000  - the shmoo group -
**
**  This program is free software; you can redistribute it and/or
**  modify it under the terms of the GNU General Public License
**  as published by the Free Software Foundation; either version 2
**  of the License, or (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
** 
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
**
**************************************************************************************/

/**************************************************************************************
**
**    The Shmoo Group (TSG)
**
**    File:      scale.c
**    Author:    Brian Wotring
**
**    Date:      January 4, 2000.
**    Project:   osiris ( UNIX ).
**
**************************************************************************************/


#include "scale.h"


/**************************************************************************************
**
**    Function: main
**
**    Purpose:  system driver for scale
**
**    Return:   zero, if all goes well.
**
**************************************************************************************/

int main( int argument_count, char *argument_list[] )
{
    init( argument_count, argument_list );

    process();
    quit();

    return 0;
}

/**************************************************************************************
**
**    Function: init 
**
**    Purpose:  check the arguments and try to open the database files
**              and output file.  If the output file exists, prompt for
**              overwrite, exit on error.
**
**    Return:   none.
**
**************************************************************************************/

void init( int argument_count, char *argument_list[]  )
{
    check_arguments( argument_count, argument_list );
    open_database_files();

    read_database_table_of_contents();
    print_database_table_of_contents();
    compare_database_table_of_contents();

    /* initialize the duplicate hash lists. */
 
    left_duplicates  = new_string_list();
    right_duplicates = new_string_list();

    /* output path exists, prompt for overwrite. */

    if( access( output_file_path, F_OK ) == 0 )
    {
        printf( "\noutput file: %s already exists, overwrite? [y/n] ", output_file_path );

	if( getc( stdin ) == 'n' )
	    log_message_exit( "...exiting" );

	if( unlink( output_file_path ) != 0 )
	    log_message_exit( "unable to overwrite old database, exiting." );
    }

    printf( "\n" );

    if( ( strlen( output_file_path ) > 0 ) && ( output_file = fopen( output_file_path, "w" ) ) == NULL )
	log_message_exit( "unable to open output path: %s", output_file_path );
}


void print_database_info( const char *database_name )
{
    entry db_entry;
    table_of_contents contents;
 
    /* verify that the database exist. */
  
    if( access( database_name, R_OK ) != 0 )
        log_message_exit( "database: %s does not exist or cannot be opened.", database_name );
  
    /* open database. */

    if( ( left_database  = open_database_for_read( database_name ) ) == NULL )
        log_message_exit( "unable to open database: %s", database_name );

    /* verify database was valid.  The database_handler initially was written with */
    /* gdbm as a database.  gdbm_key_exists() would core if you opened a file not  */
    /* made with gdbm, this is a wrapper to prevent that core.                     */

    if( !database_is_valid( left_database ) )
        log_message_exit( "database: %s not a valid database file, exiting.", database_name );

    /* get table of contents , verify key exists first. */

    if( key_exists( left_database, "\nosiriscontents\n" ) )
        db_entry.data  = get_data_for_key( left_database, "\nosiriscontents\n" );

    else
        log_message_exit( "left database corrupted or not an osiris database, exiting." );

    /* get the elements of the table of contents from the database. */

    contents.version             = get_entry_element_with_index( &db_entry, 0 );
    contents.hash_algorithm      = get_entry_element_with_index( &db_entry, 1 );
    contents.user                = get_entry_element_with_index( &db_entry, 2 );
    contents.date                = get_entry_element_with_index( &db_entry, 3 );
    contents.number_of_entries   = get_entry_element_with_index( &db_entry, 4 );
    contents.directory_list      = get_entry_element_with_index( &db_entry, 5 );

    /* print database information to console. */

    log_message( "\n[ database: %s ]\n", database_name );

    log_message( "created by:\t%s", contents.user );
    log_message( "created:\t%s", contents.date );
    log_message( "osiris version:\t%s", contents.version );
    log_message( "hash algorithm:\t%s", contents.hash_algorithm );
    log_message( "file entries:\t%s", contents.number_of_entries );

    log_message( "directories scanned:\n\n%s", contents.directory_list );
}

/**************************************************************************************
**
**    Function: read_database_table_of_contents
**
**    Purpose:  reads the table of contents from each database.  If the contents
**		is not readable, an error is printed, and scale exits.
**
**    Return:   none.
**
**************************************************************************************/

void read_database_table_of_contents()
{
    entry left_entry;
    entry right_entry;

    if( key_exists( left_database, "\nosiriscontents\n" ) )
        left_entry.data  = get_data_for_key( left_database, "\nosiriscontents\n" );

    else
	log_message_exit( "left database corrupted or not an osiris database, exiting." );


    if( key_exists( right_database, "\nosiriscontents\n" ) )
        right_entry.data = get_data_for_key( right_database, "\nosiriscontents\n" );

    else
        log_message_exit( "right database corrupted or not an osiris database, exiting." );

    left_contents.version             = get_entry_element_with_index( &left_entry, 0 );
    right_contents.version            = get_entry_element_with_index( &right_entry, 0 );

    left_contents.hash_algorithm      = get_entry_element_with_index( &left_entry, 1 );
    right_contents.hash_algorithm     = get_entry_element_with_index( &right_entry, 1 );

    left_contents.user   	      = get_entry_element_with_index( &left_entry, 2 );
    right_contents.user  	      = get_entry_element_with_index( &right_entry, 2 );

    left_contents.date   	      = get_entry_element_with_index( &left_entry, 3 );
    right_contents.date  	      = get_entry_element_with_index( &right_entry, 3 );

    left_contents.number_of_entries   = get_entry_element_with_index( &left_entry, 4 );
    right_contents.number_of_entries  = get_entry_element_with_index( &right_entry, 4 );

    left_contents.directory_list      = get_entry_element_with_index( &left_entry, 5 );
    right_contents.directory_list     = get_entry_element_with_index( &right_entry, 5 );
}

/**************************************************************************************
**
**    Function: print_database_table_of_contents
**
**    Purpose:  prints the table of contents from each database.  This will be
**		in a readable form, not the form fetched from the database.
**
**    Return:   none.
**
**************************************************************************************/

void print_database_table_of_contents()
{
    log_message( "\n[ left database ]\t\t\t\t[ right database ]\n" );

    log_message( "created by:\t%s\t\t\t\tcreated by:\t%s", left_contents.user, right_contents.user );
    log_message( "created:\t%s\tcreated:\t%s", left_contents.date, right_contents.date );

    log_message( "osiris version:\t%s\t\t\t\tosiris version:\t%s", left_contents.version, right_contents.version );

    log_message( "hash algorithm:\t%s\t\t\t\thash algorithm:\t%s", left_contents.hash_algorithm, 
                                                                   right_contents.hash_algorithm );

    log_message( "file entries:\t%s\t\t\t\tfile entries:\t%s", left_contents.number_of_entries, 
                                                               right_contents.number_of_entries );

    log_message( "" );
}

/**************************************************************************************
**
**    Function: compare_database_table_of_contents
**
**    Purpose:  gets the table of contents from each database and determines
**		if it makes sense to compare them.  If not, the user is prompted
**		and informed of their differences.
**
**		Currently, the two fields that make this difference are the hash
**		field, and the directory list field.  If these differ, scale may
**		produce crazy output ( duh ).  hash is index 0, list is index 4.
**
**    Return:   none.
**
**************************************************************************************/

void compare_database_table_of_contents()
{
    /* warn user of different hash algorithms. */

    if( strcmp( left_contents.hash_algorithm, right_contents.hash_algorithm ) != 0 )
    {
        printf( "databases were made with different hash algorithms, continue? [y/n] " );
	fflush( stdout );

        if( getc( stdin ) == 'n' )
            log_message_exit( "...exiting" );

	else
	    getc( stdin );
    }
    /* warn user of different directory lists. */

    if( strcmp( left_contents.directory_list, right_contents.directory_list ) != 0 )
    {
	
        printf( "databases were made based on a different set of directories, continue? [y/n] " );
	fflush( stdout );

        if( getc( stdin ) == 'n' )
            log_message_exit( "...exiting" );
    }
}


/**************************************************************************************
**
**    Function: print_output_file_compare_header
**
**    Purpose:  prints the header for the comparison entries of the log
**		file, only if we are in pretty mode.
**
**    Return:   none.
**
**************************************************************************************/

void print_output_file_compare_header()
{
    if( pretty_mode )
    {
	fprintf( output_file, "\n\n" );
        fprintf( output_file, "\t\t[ LEFT ]\t\t\t\t\t\t[ RIGHT ]\t\t\t\t\t[ FILE ]" );
        fprintf( output_file, "\n\n" );

	fflush( output_file );
    }
}

/**************************************************************************************
**
**    Function: print_output_file_header
**
**    Purpose:  prints the first header of the log file, only if we are in
**		pretty mode.
**
**    Return:   none.
**
**************************************************************************************/

void print_output_file_header()
{
    if( pretty_mode )
    {
	fprintf( output_file, "\n" );
        fprintf( output_file, "[ NEW/MISSING FILES, DUPLICATE HASHES ]" );
        fprintf( output_file, "\n\n" );

	fflush( output_file );
    }
}


/**************************************************************************************
**
**    Function: quit
**
**    Purpose:  close the database files, output file, and print the
**              results from scale to stdout.
**
**    Return:   none.
**
**************************************************************************************/

void quit()
{
    close_database_files();

    destroy_string_list( left_duplicates );
    destroy_string_list( right_duplicates );

    //destroy_table_of_contents( &left_contents );
    //destroy_table_of_contents( &right_contents );

    if( output_file )
        fclose( output_file );

    print_scale_results();
    print_program_info();
}


/**************************************************************************************
**
**    Function: destroy_table_of_contents
**
**    Purpose:  free memory allocated to store the table of contents.
**		for each database.
**
**    Return:   none.
**
**************************************************************************************/

void destroy_table_of_contents( table_of_contents *contents )
{
    if( contents->version )
	free( contents->version );
 
    if( contents->hash_algorithm )
	free( contents->hash_algorithm );

    if( contents->user )
	free( contents->user );

    if( contents->date )
	free( contents->date );

    if( contents->number_of_entries )
	free( contents->number_of_entries );

    if( contents->directory_list )
	free( contents->directory_list );
}

/**************************************************************************************
**
**    Function: print_scale_results
**
**    Purpose:  print all counters to stdout.
**
**    Return:   none.
**
**************************************************************************************/

void print_scale_results()
{
    log_message( "" );

    log_message( "==> %d files with different hashes.", file_hash_diff );
    log_message( "==> %d files with different modification dates.", file_date_diff );
    log_message( "==> %d files with different attributes.", file_attrib_diff );

    log_message( "" );

    log_message( "==> %d files with a different uid.", file_uid_diff );
    log_message( "==> %d files with a different gid.", file_gid_diff );
    log_message( "==> %d files with a different suid status.", file_suid_diff );

    log_message( "" );

    log_message( "==> %d files in the right database not found in the left database ( new ).", file_new_count );
    log_message( "==> %d files in the left database not found in the right database ( missing ).", file_missing_count );

    log_message( "" );
}


/************************************************************************************
**
**    Function: int process 
**
**    Purpose:  main driver for scale.  This function loops through the left
**              database checking for modified entries as compared to the right
**              database.  In addition, files added or removed from the left
**              database are logged.
**
**    Return:   none.
**
**************************************************************************************/

void process()
{
    char *key;
    char *next_key;

    entry left_entry;
    entry right_entry;

    print_output_file_header();

    check_for_missing_entries();
    check_for_new_entries();

    if( psychonazi_mode )
    {
        check_for_duplicate_left_hashes();
        check_for_duplicate_right_hashes();
    }

    print_output_file_compare_header();
    key = get_first_key( left_database );

    log_message( "==> comparing file entries..." );

    while( key )
    {
	next_key = get_next_key( left_database, key );

        /* we only compare items that are in both databases.  */
        /* also, we skip over the table of contents key when  */
        /* we get to it.                                      */

	if( !key_exists( right_database, key ) || strcmp( key, "\nosiriscontents\n" ) == 0 )
   	{
	    free( key );
	    key = next_key;

	    continue;
	}

	left_entry.data  = get_data_for_key( left_database, key );
	left_entry.key   = key;

        right_entry.data = get_data_for_key( right_database, key );
	right_entry.key  = key;

	/* perform the comparisons upon the entries. */

	compare_entry_hash( &left_entry, &right_entry );
        compare_entry_date( &left_entry, &right_entry );
  	compare_entry_attributes( &left_entry, &right_entry );

	compare_entry_uid( &left_entry, &right_entry );
	compare_entry_gid( &left_entry, &right_entry );
	compare_entry_suid( &left_entry, &right_entry );

	free( key );
        key = next_key;
    }
}


/**************************************************************************************
**
**    Function: check_for_missing_entries
**
**    Purpose:  loop through the left database.  If an entry is not found in
**              the right database, it is logged appropriatly, and counted.
**
**    Return:   none.
**
**************************************************************************************/

void check_for_missing_entries()
{
    char *key;
    char *next_key;

    log_message( "==> checking left database for missing entries..." );
    key = get_first_key( left_database );

    /* sequential step through left database. */

    while( key )
    {
	next_key = get_next_key( left_database, key );

        if( !( key_exists( right_database, key ) ) )
        {
            if( pretty_mode )
                fprintf( output_file, "missing file:\t\t%s\n", key );

            else
	        fprintf( output_file, "RMISSING\t%s\n", key );

	    file_missing_count++;
	}

	/* key returned is dynamic from handler, so we need to     */
	/* free it and get a new key.                           */

	free( key );
	key = next_key;
    }
}



/**************************************************************************************
**
**    Function: check_for_new_entries
**
**    Purpose:  loop through the right database.  If an entry is not found in
**              the left database, it is logged appropriatly, and counted.
**
**    Return:   none.
**
**************************************************************************************/

void check_for_new_entries()
{
    char *key;
    char *next_key;

    log_message( "==> checking right database for new entries..." );
    key = get_first_key( right_database );

    /* sequential step through left database. */

    while( key )
    {
	next_key = get_next_key( right_database, key );

        if( !( key_exists( left_database, key ) ) )
        {
            if( pretty_mode )
                fprintf( output_file, "new file:\t\t%s\n", key );

            else
	        fprintf( output_file, "RNEW\t\t%s\n", key );

	    file_new_count++;
	}

	/* key returned is dynamic from handler, so we need to     */
	/* free it and get a new key.                           */

	free( key );
	key = next_key;
    }
}


/**************************************************************************************
**
**    Function: check_for_duplicate_left_hashes
**
**    Purpose:  check the left database for duplicate hashes.  create a list
**		consisting of strings.  These strings will contain the names of
**		files grouped together because they have the same hash.
**
**    Return:   none.
**
**************************************************************************************/

void check_for_duplicate_left_hashes()
{
    char *hash			= NULL;
    entry file_entry;

    string_list *hashes		= NULL;
  
    char *key			= NULL;
    char *next_key		= NULL;

    log_message( "==> checking left database for duplicate hashes..." );

    key              = get_first_key( left_database );
    file_entry.data  = get_data_for_key( left_database, key );

    hashes	     = new_string_list();

    while( key )
    {
	/* skip over the contents key. */

        if( strcmp( key, "\nosiriscontents\n" ) == 0 )
        {
            next_key = get_next_key( left_database, key );
            free( key );

	    key = next_key;
            continue;
        }

	next_key = get_next_key( left_database, key );
	hash     = get_entry_element_with_index( &file_entry, 0 );

	// go through the list of hashes.

        if( item_with_name_exists( hashes, hash ) )
        {
	    append_to_item_with_name( left_duplicates, " ", hash );
	    append_to_item_with_name( left_duplicates, key, hash );
	}

	else
	{
	    add_item_with_name( hashes, NULL, hash );
	    add_item_with_name( left_duplicates, key, hash );
	}

	free( key );
	free( file_entry.data );

        key = next_key;
	file_entry.data = get_data_for_key( left_database, key );
    }

    destroy_string_list( hashes );
    log_duplicate_left_hashes();
}


/**************************************************************************************
**
**    Function: check_for_duplicate_right_hashes
**
**    Purpose:  check the right database for duplicate hashes.  create a list
**		consisting of strings.  These strings will contain the names of
**		files grouped together because they have the same hash.
**
**    Return:   none.
**
**************************************************************************************/

void check_for_duplicate_right_hashes()
{
    char *hash			= NULL;
    entry file_entry;

    string_list *hashes		= NULL;
  
    char *key			= NULL;
    char *next_key		= NULL;

    log_message( "==> checking left database for duplicate hashes..." );

    key              = get_first_key( right_database );
    file_entry.data  = get_data_for_key( right_database, key );

    hashes	     = new_string_list();

    while( key )
    {
        /* skip over the contents key. */

        if( strcmp( key, "\nosiriscontents\n" ) == 0 )
        {
            next_key = get_next_key( right_database, key );
            free( key );

            key = next_key;
            continue;
        }

	next_key = get_next_key( right_database, key );
	hash     = get_entry_element_with_index( &file_entry, 0 );

	// go through the list of hashes.

        if( item_with_name_exists( hashes, hash ) )
        {
	    append_to_item_with_name( right_duplicates, " ", hash );
	    append_to_item_with_name( right_duplicates, key, hash );
	}

	else
	{
	    add_item_with_name( hashes, NULL, hash );
	    add_item_with_name( right_duplicates, key, hash );
	}

	free( key );
	free( file_entry.data );

        key = next_key;
	file_entry.data = get_data_for_key( right_database, key );
    }

    destroy_string_list( hashes );
    log_duplicate_right_hashes();
}

/**************************************************************************************
**
**    Function: log_duplicate_left_hashes
**
**    Purpose:  go through the duplicate list.  If any string has more than one
**		space in it, it means there is more than one file in it, meaning
**		there are duplicates files, so log to the appropriate devices.
**
**    Return:   none.
**
**************************************************************************************/

void log_duplicate_left_hashes()
{
    int index;

    for( index = 0; index < left_duplicates->size; index++ )
    {
	if( strstr( left_duplicates->list[index]->data, " " ) )
	{
	    /* write to output file, format may vary if in pretty mode. */

	    if( pretty_mode )
                fprintf( output_file, "duplicate left hash\t\t%s\n", left_duplicates->list[index]->data );

	    else
	        fprintf( output_file, "MULTILH\t\t%s\n", left_duplicates->list[index]->data );

	    left_duplicate_count++;
	}
    }
}

/**************************************************************************************
**
**    Function: log_duplicate_right_hashes
**
**    Purpose:  go through the duplicate list.  If any string has more than one
**		space in it, it means there is more than one file in it, meaning
**		there are duplicates files, so log to the appropriate devices.
**
**    Return:   none.
**
**************************************************************************************/

void log_duplicate_right_hashes()
{
    int index;

    for( index = 0; index < right_duplicates->size; index++ )
    {
	if( strstr( right_duplicates->list[index]->data, " " ) )
	{
	    /* write to output file, format may vary if in pretty mode. */

	    if( pretty_mode )
                fprintf( output_file, "duplicate right hash\t\t%s", right_duplicates->list[index]->data );

	    else
	        fprintf( output_file, "MULTIRH\t\t%s\n", right_duplicates->list[index]->data );

	    right_duplicate_count++;
	}
    }
}


/**************************************************************************************
**
**    Function: compare_entry_hash
**
**    Purpose:  the checksums of the left and right entry are compared.
**              If they are unequal, they are logged appropriatly, and counted.
**
**    Return:   none.
**
**************************************************************************************/

void compare_entry_hash( entry *left_entry, entry *right_entry )
{
    char *left_checksum;
    char *right_checksum;

    char *file_path;

    if( left_entry == NULL || right_entry == NULL )
	return;

    left_checksum  = get_entry_element_with_index( left_entry, 0 );
    right_checksum = get_entry_element_with_index( right_entry, 0 );

    file_path = left_entry->key;

    /* checksums not equal, log to file, maybe to console. */

    if( strcmp( left_checksum, right_checksum ) != 0 )
    {
	/* write to output file, format may vary if in pretty mode. */

	if( pretty_mode )
            fprintf( output_file, "hash:\t\t%s\thash:\t\t%s\t%s\n", left_checksum, right_checksum, file_path );

	else
	    fprintf( output_file, "HASH\t\t%s\t%s\t\t%s\n", left_checksum, right_checksum, file_path );

	file_hash_diff++;
    }

    free( left_checksum );
    free( right_checksum );
}


/**************************************************************************************
**
**    Function: compare_entry_date
**
**    Purpose:  the modification dates of the left and right entry are compared.
**              If they are unequal, they are logged appropriatly, and counted.
**
**    Return:   none.
**
**************************************************************************************/\

void compare_entry_date( entry *left_entry, entry *right_entry )
{
    char *left_date;
    char *right_date;

    char *file_path;

    if( left_entry == NULL || right_entry == NULL )
	return;

    left_date  = get_entry_element_with_index( left_entry, 1 );
    right_date = get_entry_element_with_index( right_entry, 1 );

    file_path = left_entry->key;

    /* checksums not equal, log to file, maybe to console. */

    if( strcmp( left_date, right_date ) != 0 )
    {
	/* write to output file, format may vary if in pretty mode. */

	if( pretty_mode )
            fprintf( output_file, "mod date:\t%s\t\tmod date:\t%s\t\t%s\n", left_date, right_date, file_path );

	else
	    fprintf( output_file, "DATE\t\t%s\t\t%s\t\t\t%s\n", left_date, right_date, file_path );

	file_date_diff++;
    }

    free( left_date );
    free( right_date );
}


/**************************************************************************************
**
**    Function: compare_entry_attributes
**
**    Purpose:  the permissions attributes of the left and right entry are compared.
**              If they are unequal, they are logged appropriatly, and counted.
**
**    Return:   none.
**
**************************************************************************************/

void compare_entry_attributes( entry *left_entry, entry *right_entry )
{
    char *left_attributes;
    char *right_attributes;

    char *file_path;

    if( left_entry == NULL || right_entry == NULL )
	return;

    left_attributes  = get_entry_element_with_index( left_entry, 2 );
    right_attributes = get_entry_element_with_index( right_entry, 2 );

    file_path = left_entry->key;

    /* attributes not equal, log to file, maybe to console. */

    if( strcmp( left_attributes, right_attributes ) != 0 )
    {
	/* write to output file, format may vary if in pretty mode. */

	if( pretty_mode )
            fprintf( output_file, "attributes:\t%s\t\t\t\tattributes:\t%s\t\t\t\t%s\n", left_attributes, right_attributes, file_path );

	else
	    fprintf( output_file, "ATTRIBUTES\t%s\t\t%s\t\t\t%s\n", left_attributes, right_attributes, file_path );

	file_attrib_diff++;
    }

    free( left_attributes );
    free( right_attributes );
}

/**************************************************************************************
**
**    Function: get_entry_element_with_entry_index
**
**    Purpose:  used by the compare functions to extract a field element from
**              an entry.  elements are indexed starting at zero and are delimited
**              by semicolons.  The format is as follows:
**
**              "hash;date;attributes;uid;gid;suid"
**
**    Return:   pointer to element value, dynamic allocation ( use carefully ).
**
**************************************************************************************/

char * get_entry_element_with_index( entry *file_entry, unsigned int index )
{
    int count		= 0;
    int index_count	= 0;
    int element_size	= 0;

    char *element;

    if( file_entry == NULL || file_entry->data == NULL )
        return 0;

    /* element could never, realistically, be bigger than entry. */

    element = safemalloc( sizeof( char ) * strlen( file_entry->data ) );

    /* loop through entry, if index is found, copy data  */
    /* until the index stops.                            */
 
    for( count = 0; count < strlen( file_entry->data ); count++ )
    {
        if( file_entry->data[count] == ';' )
        {
	    count++;
            index_count++;
        }

	if( index_count == index )
	    element[element_size++] = file_entry->data[count];    
    }

    element[element_size] = '\0';
    return element;
}

/**************************************************************************************
**
**    Function: compare_entry_uid
**
**    Purpose:  the uid values of the left and right entry are compared.
**              If they are unequal, they are logged appropriatly, and counted.
**
**    Return:   none.
**
**************************************************************************************/

void compare_entry_uid( entry *left_entry, entry *right_entry )
{
    char *left_uid;
    char *right_uid;

    char *file_path;

    if( left_entry == NULL || right_entry == NULL )
	return;

    left_uid  = get_entry_element_with_index( left_entry, 3 );
    right_uid = get_entry_element_with_index( right_entry, 3 );

    file_path = left_entry->key;

    /* attributes not equal, log to file, maybe to console. */

    if( strcmp( left_uid, right_uid ) != 0 )
    {
	/* write to output file, format may vary if in pretty mode. */

	if( pretty_mode )
            fprintf( output_file, "uid:\t\t%s\t\t\t\t\tuid:\t\t%s\t\t\t\t\t%s\n", left_uid, right_uid, file_path );

	else
	    fprintf( output_file, "UID\t\t%s\t\t\t\t\t%s\t\t\t\t\t\t%s\n", left_uid, right_uid, file_path );

	file_uid_diff++;
    }

    free( left_uid );
    free( right_uid );
}


/**************************************************************************************
**
**    Function: compare_entry_gid
**
**    Purpose:  the gid values of the left and right entry are compared.
**              If they are unequal, they are logged appropriatly, and counted.
**
**    Return:   none.
**
**************************************************************************************/

void compare_entry_gid( entry *left_entry, entry *right_entry )
{
    char *left_gid;
    char *right_gid;

    char *file_path;

    if( left_entry == NULL || right_entry == NULL )
	return;

    left_gid  = get_entry_element_with_index( left_entry, 4 );
    right_gid = get_entry_element_with_index( right_entry, 4 );

    file_path = left_entry->key;

    /* attributes not equal, log to file, maybe to console. */

    if( strcmp( left_gid, right_gid ) != 0 )
    {
	/* write to output file, format may vary if in pretty mode. */

	if( pretty_mode )
            fprintf( output_file, "gid:\t\t%s\t\t\t\t\tgid:\t\t%s\t\t\t\t\t%s\n", left_gid, right_gid, file_path );

	else
            fprintf( output_file, "GID\t\t%s\t\t\t\t\t%s\t\t\t\t\t\t%s\n", left_gid, right_gid, file_path );

	file_gid_diff++;
    }

    free( left_gid );
    free( right_gid );
}


/**************************************************************************************
**
**    Function: compare_entry_suid
**
**    Purpose:  the suid values of the left and right entry are compared.
**              If they are unequal, they are logged appropriatly, and counted.
**
**    Return:   none.
**
**************************************************************************************/

void compare_entry_suid( entry *left_entry, entry *right_entry )
{
    char *left_suid;
    char *right_suid;

    char *file_path;

    if( left_entry == NULL || right_entry == NULL )
	return;

    left_suid  = get_entry_element_with_index( left_entry, 5 );
    right_suid = get_entry_element_with_index( right_entry, 5 );

    file_path = left_entry->key;

    /* attributes not equal, log to file, maybe to console. */

    if( strcmp( left_suid, right_suid ) != 0 )
    {
	/* write to output file, format may vary if in pretty mode. */

	if( pretty_mode )
            fprintf( output_file, "suid:\t\t%s\t\t\t\t\tsuid:\t\t%s\t\t\t\t\t%s\n", left_suid, right_suid, file_path );

	else
            fprintf( output_file, "SUID\t\t%s\t\t\t\t\t%s\t\t\t\t\t\t%s\n", left_suid, right_suid, file_path );

	file_suid_diff++;
    }

    free( left_suid );
    free( right_suid );
}


/**************************************************************************************
**
**    Function: void check_arguments
**
**    Purpose:  parses command line arguments.
**
**    Return:   none.
**
**************************************************************************************/

void check_arguments( int argument_count, char *argument_list[] )
{
    int count;

    if( argument_count < 3 )
    {
        print_usage();
	exit(0);
    }

    for( count = 1; count < argument_count; count++ )
    {
        /* get database paths. */

	if( strcmp( argument_list[count], "-l" ) == 0  && ( count < ( argument_count - 1 ) ) )
	    strcpy( left_database_path, argument_list[++count] );

	else if( strcmp( argument_list[count], "-r" ) == 0  && ( count < ( argument_count - 1 ) ) )
	    strcpy( right_database_path, argument_list[++count] );


	else if( strcmp( argument_list[count], "-p" ) == 0 && ( count < ( argument_count - 1 ) ) )
        {
	    print_database_info( argument_list[++count] );
	    exit(0);
	}

 	else if( strcmp( argument_list[count], "-o" ) == 0 && ( count < ( argument_count - 1 ) ) )
            strcpy( output_file_path, argument_list[++count] );	    

        else if( strcmp( argument_list[count], "-n" ) == 0 )
	    pretty_mode = 1;

        else if( strcmp( argument_list[count], "-psychonazi" ) == 0 )
	    psychonazi_mode = 1;

	else
        {
	    print_usage();
	    exit(0);
	}
    }

    if( strlen( output_file_path ) == 0 )
    {
	print_usage();
	log_message_exit( "error: must specify an output file path." );
    }
}



/**************************************************************************************
**
**    Function: void open_database_files
**
**    Purpose:  verifies the databases have valid file names, if so, checks
**              for read permissions on them, if so, they are opened and 
**              checked for succesfull open.
**
**    Return:   none.
**
**************************************************************************************/

void open_database_files()
{
    /* verify paths. */

    if( strlen( left_database_path ) == 0 )
	log_message_exit( "no left database specified." );

    if( strlen( right_database_path ) == 0 )
	log_message_exit( "no right database specified." );

    /* verify that the databases exist. */

    if( access( left_database_path, R_OK ) != 0 )
        log_message_exit( "left database: %s does not exist or cannot be opened.", left_database_path );

    if( access( right_database_path, R_OK ) != 0 )
        log_message_exit( "right database: %s does not exist or cannot be opened.", right_database_path );
 
    /* open databases. */

    left_database  = open_database_for_read( left_database_path );
    right_database = open_database_for_read( right_database_path );

    /* exit if either database was not to be opened. */

    if( left_database == NULL )
        log_message_exit( "unable to open left database: %s", left_database_path );

    if( right_database == NULL )
        log_message_exit( "unable to open right database: %s", right_database_path );

    /* verify database was valid.  The database_handler initially was written with */
    /* gdbm as a database.  gdbm_key_exists() would core if you opened a file not  */
    /* made with gdbm, this is a wrapper to prevent that core.                     */

    if( !database_is_valid( left_database ) )
        log_message_exit( "left database: %s not a valid database file, exiting.", left_database_path );

    if( !database_is_valid( right_database ) )
	log_message_exit( "right database: %s not a valid database file, exiting.", right_database_path );
}


/**************************************************************************************
**
**    Function: void close_database_files
**
**    Purpose:  tries to close the database files.
**
**    Return:   none.
**
**************************************************************************************/

void close_database_files()
{
    log_message_debug( "closing database." );

    close_database( left_database );
    close_database( right_database );

    free( left_database );
    free( right_database );
}


/**************************************************************************************
**
**    Function: void print_usage
**
**    Purpose:  prints usage statement and command line arguments for
**		using scale.
**
**    Return:   none.
**
**************************************************************************************/

void print_usage()
{
    log_message( "" );
    log_message( "usage: scale -l <path> -r <path> -o <path> [-psychonazi] [-h] [-n] | -p <database>" );
    log_message( "" );

    log_message( "     -l               specifies left database to use in comparison." );
    log_message( "     -r               specifies right database to use in comparison." );
    log_message( "     -o               path to export the comparison results. ( e.g., /var/log/osiris.log )." );
    log_message( "     -n               print output to file in nice, human readable format." );
    log_message( "     -h               display these messages." );
    log_message( "     -p               print database table of contents and exit." );
    log_message( "     -psychonazi      report files found throughout the system with the same checksum." );

    print_program_info();
}

/**************************************************************************************
**
**    Function: void print_program_info
**
**    Purpose:  prints program name, version, authors, etc.
**
**    Return:   none.
**
**************************************************************************************/

void print_program_info()
{
    log_message( "" );

    log_message( "-------------------------------------------" );
    log_message( "  %s  - version %s                         ", program_name, program_version );
    log_message( "  www.shmoo.com                 (c) 2000   " );
    log_message( "-------------------------------------------" );

    log_message( "" );
}

